/*====================================================================================================*
* 2-Lane Drag Race Timer for 1/10 Scale RC cars                             Version 4c - 08 Mar. 2023 *
* Created by: Bill van Leeuwen (email: Billv923@outlook.com) for the Arduino UNO microcontroller      *
* Version 4c provides the following features:                                                         *
*   1. Christmas Tree Control including:                                                              *
*      a. Selectable Standard or Pro light sequencing                                                 *
*      b. Stage Indicators                                                                            *
*      b. Disqualification (Foul) red-light indicators                                                *
*   2. Lane Sensors at the following positions:                                                       *
*      a. Stage sensors (Note: Pre-stageing not supported in this version)                            *
*      b. Guard sensors (Used to determine when car leaves the start line)                            *
*      c. Trap sensors (Used in conjuction with finish line sensors to calculate speed                *
*      d. Finish Line sensors                                                                         *
*   3. Race Result Numeric Displays                                                                   *
*      a. Reaction time (Acurate to 0.001 seconds)                                                    *
*      b. Speed in mph or km/h (Selectable in code setup)                                             *
*      c. Elapsed Time (ET) (Acurate to 0.001 seconds)                                                *
*   4. Timer control via hand-held pushbutton                                                         *
*   5. Winner indication on corresponding Xmas Tree green lamp indicators                             *
*   6. Sensor fault (not ready) indication & reporting via blinking red-light indication              *
*   7. User selectable Standard or Pro Xmas tree sequencing                                           *
*   8. User selectable Auto or Manual Start                                                           *
*   9. User selectable 1 or 2 lane operation                                                          *
*                                                                                                     *
* This version uses a pair of serially cascaded TPIC6B595 power shift registers to drive the high     *
* power LED Xmas tree lights. Each light is assigned a bit in a 8-bit byte as follows:                *
*   Bit 7(MSB) - Pre-Stage Indicator (Not Used in this version)                                       *
*   Bit 6      - Stage Indicator                                                                      *
*   Bit 5      - Amber1 Indicator                                                                     *
*   Bit 4      - Amber2 Indicator                                                                     *
*   Bit 3      - Amber3 Indicator                                                                     *
*   Bit 2      - Green GO Indicator                                                                   *
*   Bit 1      - Red DQ Indicator                                                                     *
*   Bit 0(LSB) - Not used                                                                             *
* The two 8-bit bytes (1 each for lane1 and lane2) are combined into a 16-bit word that is serially   *
* sent to the TPIC6B595 power shift registers via the SPI interface (Pins D11 & D13)                  *
*                                                                                                     *
* This program is designed to work either standalone or in conjunction with the author's race         *
* management/display software running on a PC attached to the Arduino timer via the USB serial port.  *
* This release has been customized to include software to drive six Adafruit 4-digit numeric displays *
* (3 per lane) that display the racer's reaction time, speed (in mph or km/h) and elapsed time (ET).  *
*                                                                                                     *
* This program measures the start reaction time, trap sensor time and finish time for 2 lanes. The    *
* trap sensor time is used to calculate the vehicle speed during the last segment of the track.       *
* Race results are displayed on the numeric displays and optionally sent to the serial port to        *
* interface with the race management software running on the PC. Times are measured from activation   *
* of the green Xmas tree light to activation of the various sensors of each lane. Each racer has 10   * 
* seconds to finish the race. If a racer doesn't cross the finish line within that time, it is        *
* assumed that the racer had a problem or the lane was not used. In either case, the value 9.999 is   *
* assigned to the lane time as an error indication.                                                   *
*                                                                                                     *
* NOTE: This program can be set to calculate & display speed in either miles-per-hour (mph) or        *
* kilometers-per-hour (km/h). When set to display speed in mph the trap & finish line sensor          *
* seperation must be 5.28 feet. When used to display speed in km/h the trap & finish line sensor      *
* seperation must be 2 meters. A variable named "kmh_Flag" is set during code setup  to determine     *
* whether the speed is calculated in mph or km/h.                                                     *
*                                                                                                     *
* A hand-held pushbutton switch is provided to remotely start the Xmas Tree countdown sequence or to  *
* reset the timer for the next race. Lane times and track status is sent via the USB port to a PC if  *
* connected (optional) running the race management/display software.                                  *
*                                                                                                     *
* Refer to the Schematic Diagram, parts list and design notes that accompany this software for more   *
* information. Feel free to contact the author at Billv923@outlook.com if any questions               *
*====================================================================================================*/
//include the SPI library:
#include <SPI.h>                     //Serial Peripheral Interface library to drive the Xmas tree LEDs
#include <Wire.h>                    //Library used to establish the I2C serial interface
#include <Adafruit_LEDBackpack.h>    //Library used with Adafruit 4-digit numeric displays
#include <Adafruit_GFX.h>            //Used to support Adafruit_LEDBackpack.h library

//Initialization
//Sensor logic state definitions (Can be reversed depending on sensor used)
#define Tripped       1              //Sensor logic state when tripped (0 or 1)

//Pin Assignments
//NOTE1: MOSI pin 11 & SCK pin 13 are automatically assigned by SPI Library)
//NOTE2: I2C Interface Serial Data (SDA) line & Serial Clock (SCL) lines (Pins A4 & A5) are
//       automatically assigned in the Wire.h library
#define Stg_Sense1    2              //Lane1 Stage Sensor - Digital Input Pin D2
#define Stg_Sense2    3              //Lane2 Stage Sensor - Digital Input Pin D3
#define Grd_Sense1    4              //Lane1 Guard Sensor - Digital Input Pin D4
#define Grd_Sense2    5              //Lane2 Guard Sensor - Digital Input Pin D5
#define Trap_Sense1   6              //Lane1 Trap Sensor - Digital Input Pin D4
#define Trap_Sense2   7              //Lane2 Trap Sensor - Digital Input Pin D5
#define Fin_Sense1    8              //Finish Sensor Lane1 - Digital Input Pin D6
#define Fin_Sense2    9              //Finish Sensor Lane2 - Digital Input Pin D7
#define OutReg_CLK   10              //TPIC6B595 Output Register Clock
#define hand_sw      12              //Manual hand-held switch - Digital Input Pin D12
#define XTmode_sw    A0              //Xmas tree Std/Pro mode select switch - Digital Input Pin A0 (14)
#define AutoMan_sw   A1              //Timer Auto/Manual mode select switch - Digital Input Pin A1 (15)
#define Racers_sw    A2              //Xmas tree Std/Pro mode select switch - Digital Input Pin A2 (16)


//Variable Initialization
enum racestate {NotReady, Ready, Staged, Countdown, Racing, Finished};
byte racestate;                      //0=NotReady, 1=Ready, 2=Countdown, 3=Racing, 4=Finished
const byte Stage = B01000000;        //Stage indicator bit 6
const byte Amber1 = B00100000;       //Amber1 indicator bit 5
const byte Amber2 = B00010000;       //Amber2 indicator bit 4
const byte Amber3 = B00001000;       //Amber3 indicator bit 3
const byte Green = B00000100;        //Green GO indicator bit 2
const byte Disq = B00000010;         //Red disqualified indicator bit 1
const byte AmberAll = B00111000;     //Amber1,2,& 3 indicator bits 3, 4 & 5
boolean Lane1_ready;                 //False if any Lane1 sensor is obstructed
boolean Lane2_ready;                 //False if any Lane2 sensor is obstructed
boolean Track_ready;                 //False if track is not ready
boolean Stg1_Flag;                   //True if Lane1 Stage sensor was tripped
boolean Stg2_Flag;                   //True if Lane2 Stage sensor was tripped
boolean Grd1_Flag;                   //True if Lane1 Guard sensor was tripped
boolean Grd2_Flag;                   //True if Lane2 Guard sensor was tripped
boolean Trap1_Flag;                  //True if Lane1 Trap sensor was tripped
boolean Trap2_Flag;                  //True if Lane2 Trap sensor was tripped
boolean L1blink_Flag;                //True if Lane1 Stage indicator is commanded to blink
boolean L2blink_Flag;                //True if Lane2 Stage indicator is commanded to blink
boolean L1stage_Flag;                //True if Lane1 vehicle is staged
boolean L2stage_Flag;                //True if Lane2 vehicle is staged
boolean Hold_Flag;                   //True if 'Hold' command rec'd from PC to disable start switch
boolean ProTree_Flag;                //Where: false=Standard "Sportsmans" mode, true="Pro" mode
boolean Autostart_Flag;              //Where: false='Manual start', True='Auto-start'
boolean OneRacer_Flag;               //Where: false='2 racers', true='1 racer'
boolean Win_Flag[2];                 //Where: true=winner
byte Lane1_lights;                   //Lane1 combined indicator bits
byte Lane2_lights;                   //Lane2 combined indicator bits
int i,n;                             //Working variables
int lane_place[2];                   //Lane finish place array
int Seq_State;                       //Progression state of Xmas Tree sequencing lights (0,1,2,3,4,5)
                                     //Where: 0=off, 1=Amber1, 2=Amber2, 3=Amber3, 4=Green, 5=Red (Disq)
int lane;                            //Lane number
int finish_order;                    //Finish order
int Num_Vehicles;                    //Number of vehicles racing (active lanes): 1 or 2
int fincount;                        //Tallies racers crossing finish line to determine race over
int Foul_Flag[2];                    //Equals 1 if 1'st to foul, 2 if 2'nd to foul
unsigned long max_duration = 10000;  //Default race duration in milliseconds
unsigned long Start_time;            //Time when green light illuminates (milliseconds)
unsigned long Exp_Time;              //Expiration Time (milliseconds)
unsigned long Clk_Time;              //CPU Clock Time (milliseconds)
unsigned long Xtime;                 //Multi-use time variable (milliseconds)
unsigned long Seq_Time;              //CPU clock time variable used in Xtree Sequence routine
unsigned long React_time[2];         //Reaction time data array (milliseconds)
unsigned long Trap_time[2];          //Trap time data array (milliseconds)
unsigned long Fin_time[2];           //Elapsed finish time data array (milliseconds)
unsigned long blinkTime;             //Time variable used in winning lane blink routine
float TrapFin_Separation = 5.28;     //Trap / Finish Sensor separation in feet
float Time;                          //Working variable - Calculated Lane Time (seconds)
float Temp;                          //Working temporary floating point variable
float Rtime[2];                      //Reaction time data array (seconds)
float Etime[2];                      //Elapsed time data array (seconds)
float Otime[2];                      //Overall time data array (seconds)
float Speed[2];                      //Speed data array (mph)
char SMsg0[4] = "NRD";               //Serial message - Not Ready
char SMsg1[4] = "RDY";               //Serial message - Ready
char SMsg2[4] = "STG";               //Serial message - Staged
char SMsg3[4] = "CNT";               //Serial message - Countdown Sequence underway
char SMsg4[4] = "RAC";               //Serial message - Racing
char SMsg5[4] = "FIN";               //Serial message - Finished
char SMsg7[4] = "TRK";               //Serial message - Track Sensor(s) blocked
char SMsg8[4] = "RST";               //Serial message - Manual RESET executed
String error_codes = "";             //Message string populated with error codes
String serial_data = "";             //Serial data input string
String Str1;                         //String variable

//USER SELECTABLE SETTINGS
boolean kmh_Flag = false;            //Set to true to display speed in km/h
int dispBrightness = 10;             //Range 0 - 15 (where: 0=dim, 15=max brightness)

//Adafruit library 4-digit display definitions (6 displays)
int Adafruit_Address[6]={0x70, 0x71, 0x72, 0x73, 0x74, 0x75};  //I2C Adafruit Display Addresses
Adafruit_7segment LED_Disp[6];  //Define "LED_Disp[x]" as the Adafruit 7-segment display name

//======================================= SETUP ROUTINE =======================================
void setup() {  
//Hardware Setup
  //Initialize track sensors and Reset switch pins as inputs
  //Note: Sense inputs transition from LOW to HIGH when light beam is interrupted
  pinMode(Stg_Sense1, INPUT_PULLUP);               //Lane1 Stage sensor
  pinMode(Stg_Sense2, INPUT_PULLUP);               //Lane2 Stage sensor
  pinMode(Grd_Sense1, INPUT_PULLUP);               //Lane1 Stage sensor
  pinMode(Grd_Sense2, INPUT_PULLUP);               //Lane2 Stage sensor
  pinMode(Trap_Sense1, INPUT_PULLUP);              //Lane1 Trap sensor
  pinMode(Trap_Sense2, INPUT_PULLUP);              //Lane2 Trap sensor
  pinMode(Fin_Sense1, INPUT_PULLUP);               //Lane1 Finish line sensor
  pinMode(Fin_Sense2, INPUT_PULLUP);               //Lane2 Finish line sensor
  pinMode(hand_sw, INPUT_PULLUP);                  //Switch is normally HIGH & transitions to LOW when pressed
  pinMode(OutReg_CLK, OUTPUT);                     //Set Pin 10 as digital output
  pinMode(XTmode_sw, INPUT_PULLUP);                //Open = Standard "Sportsmans" mode, Closed = "Pro" mode
  pinMode(AutoMan_sw, INPUT_PULLUP);               //Open = Manual mode, Closed = Auto mode
  pinMode(Racers_sw, INPUT_PULLUP);                //Open = 2 racers, Closed = 1 racer
  digitalWrite(OutReg_CLK, LOW);                   //Set default state
  
//Software Setup
  //Initialize Serial Peripheral Interface (SPI) to drive the Xmas Tree lights
  SPI.beginTransaction(SPISettings(15000000, LSBFIRST, SPI_MODE3));
  SPI.begin();                                     //Start SPI
  Lane1_lights = 0x00;                             //Set all Lane1 lights off
  Lane2_lights = 0x00;                             //Set all Lane2 lights off
  Set_Lights(Lane1_lights, Lane2_lights);          //Call subroutine to set Xmas tree lights

  //Initialize the Adafruit 7-segment displays
  for (i=0; i<6; i++) {                            //Loop through all 6 displays
    LED_Disp[i] = Adafruit_7segment();             //Identify display as Adafruit 7-segment display
    LED_Disp[i].begin(Adafruit_Address[i]);        //Assign appropriate display I2C address
    LED_Disp[i].clear();                           //Set the display clear command
    LED_Disp[i].drawColon(false);                  //Set the colon off command
    LED_Disp[i].writeDisplay();                    //And send above commands to the display
    LED_Disp[i].setBrightness(dispBrightness);     //Set display intensity (0 - 15)
  }
  
  Serial.begin(9600);                              //Enable serial comms to PC at 9600 baud
  racestate = NotReady;                            //Set initial race state to not ready
  Foul_Flag[0] = 0;                                //Preset Lane1 disqualification flag to false
  Foul_Flag[1] = 0;                                //Preset Lane2 disqualification flag to false
  ProTree_Flag=false;                              //Preset Xmas Tree functionality to "Sportsman" mode
  Autostart_Flag=false;                            //Preset Autostart flag to "manual" start
  OneRacer_Flag=false;                             //Preset one racer flag to false
  Num_Vehicles = 2;                                //Preset number of vehicles racing (active lanes)to 2
  Serial.println("@");                             //Send acknowledge message to PC
  Lamptest();                                      //Call Xmas Tree Lamp Test subroutine
  timer_reset();                                   //Call subroutine to reset the timer
}

//======================================== MAIN LOOP ============================================
void loop() {
  //READ & PROCESS INCOMING SERIAL DATA
  serial_data = get_serial_data();                   //Get any received serial data
  if (serial_data.indexOf("R") >= 0) {               //Check for Reset request from PC
    timer_reset();                                   //Call subroutine to reset the timer
  }
  if (serial_data.indexOf("S") >= 0) {               //Check for race start request from PC
    if (racestate == Staged) {                       //Check if both vehicles are staged
      racestate = Countdown;                         //Set race state to Countdown mode
      Seq_State = 1;                                 //Set Sequence State count to 1 (Amber1 on)
      Serial.println(SMsg3);                         //Send Countdown state message to PC
    }
  }
  if (serial_data.indexOf("H") >= 0) {               //Check for hold request from PC
    Hold_Flag = true;                                //Set Hold flag
    delay(10);                                       //Wait a bit before we send Acknowledge
    Serial.println("@");                             //Send acknowledge message to PC
  }
  if (serial_data.indexOf("C") >= 0) {               //Check for continue request from PC
    Hold_Flag = false;                               //Clear Hold flag
    delay(10);                                       //Wait a bit before we send Acknowledge
    if (kmh_Flag) Str1 = "@-kmh";                    //Build acknowledge string with speed setting for km/h
      else Str1 = "@-mph";                           //Build acknowledge string with speed setting for MPH
    Serial.println(Str1);                            //Send acknowledge message to PC
  }
  if (serial_data.indexOf("T") >= 0) {               //Check for Xmas tree lamp test request from PC
    Lamptest();                                      //Call Xmas Tree Lamp Test subroutine
    Serial.println("@");                             //Send Acknowledge message to PC
  }
  if (serial_data.indexOf("D") >= 0) {               //Check for debug request from PC
    Serial.println("DEBUG RESULTS");                 //Print Debug header
    Serial.print("  Pro Mode: ");                    //Print Pro Mode true/false status
    if (ProTree_Flag) Serial.println("True");
      else Serial.println("False");
    Serial.print("  Auto Start: ");                  //Print Auto Start true/false status
    if (Autostart_Flag) Serial.println("True");
      else Serial.println("False");
    Serial.print("  One Racer: ");                   //Print One Racer true/false status
    if (OneRacer_Flag) Serial.println("True");
      else Serial.println("False");
    Serial.print("  Race state: ");                  //Print race state status
    Serial.println(racestate);
    Serial.print("  Speed Setting: ");               //Print speed flag setting
    if (kmh_Flag) Serial.println("km/h");
      else Serial.println("mph");
    Serial.println("");
  }

  //CHECK & PROCESS SWITCH STATES
  if (digitalRead(XTmode_sw)==0) ProTree_Flag=true;  //Process Xmas Tree mode switch
    else ProTree_Flag=false;
  if (digitalRead(AutoMan_sw)==0) Autostart_Flag=true; //Process Auto/Manual timing mode switch
    else Autostart_Flag=false;
  if (digitalRead(Racers_sw)==0)   {                 //Check state of 1 or 2 racers switch
    OneRacer_Flag=true;                              //Set 1 racer mode if true
    Num_Vehicles = 1;                                //Set number of vehicles (active lanes) to 1
  }
  else {
    OneRacer_Flag=false;                             //Set 2 racer mode if false
    Num_Vehicles = 2;                                //Set number of vehicles (active lanes) to 2
  }

  //CHECK & PROCESS LANE 1 & 2 STAGE SENSORS
  if (digitalRead(Stg_Sense1)==Tripped) {          //Is Lane1 Stage sensor tripped?
    Lane1_lights = Lane1_lights | Stage;           //Set Lane1 stage light
    Stg1_Flag = true;                              //Set Lane1 as staged
  }
  else {
    Lane1_lights = Lane1_lights & ~Stage;          //Set Lane1 stage light to off
    Stg1_Flag = false;                             //Set Lane1 as not staged
  }
  if (digitalRead(Stg_Sense2)==Tripped) {          //Is Lane2 Stage sensor tripped?
    Lane2_lights = Lane2_lights | Stage;           //Set Lane2 stage light
    Stg2_Flag = true;                              //Set Lane2 as staged
  }
  else {
    Lane2_lights = Lane2_lights & ~Stage;          //Set Lane2 stage light to off
    Stg2_Flag = false;                             //Set Lane2 as not staged
  }
  //CHECK & PROCESS LANE 1 & 2 GAURD SENSORS
  if (digitalRead(Grd_Sense1)==Tripped) Grd1_Flag = true;  //Set Lane1 Guard flag if sensor tripped
    else Grd1_Flag = false;                        //Or set Lane1 Guard flag as not tripped
  if (digitalRead(Grd_Sense2)==Tripped) Grd2_Flag = true;  //Set Lane2 Guard flag if sensor tripped
    else Grd2_Flag = false;                        //Or set Lane2 Guard flag as not tripped
  if (racestate==Ready || racestate==Staged) {   //If in Ready or Staged state
    if (Grd1_Flag) Lane1_lights = Lane1_lights | Disq; //Set Lane1 DQ light to on
      else Lane1_lights = Lane1_lights & ~Disq;    //Set Lane1 DQ light to off
    if (Grd2_Flag) Lane2_lights = Lane2_lights | Disq; //Set Lane2 DQ light to on
      else Lane2_lights = Lane2_lights & ~Disq;    //Set Lane2 DQ light to off
  }
  
  //READY STATE
  if (racestate == Ready) {                          //Do if in the Ready race state
    if (OneRacer_Flag) {                             //Do if only 1 car racing on Lane1
      if (Stg1_Flag) {                               //Is Lane1 vehicle staged?
        delay(250);                                  //Wait 1/4 second
        if (digitalRead(Stg_Sense1)==Tripped) {      //Is Lane1 vehicle still staged?
          racestate = Staged;                        //Set race state to Staged if so
          L1stage_Flag = true;                       //Set the Lane1 staged.
          Serial.println(SMsg2);                     //Send Staged state message to PC
        }
      }
    }
    else {                                           //Do if 2 cars are racing
      if (Stg1_Flag && Stg2_Flag) {                  //Are both vehicles staged?
        delay(250);                                  //Wait 1/4 second
        if (digitalRead(Stg_Sense1)==Tripped && digitalRead(Stg_Sense2)==Tripped) { //Are both vehicles still staged?
          racestate = Staged;                        //Set race state to Staged if so
          L1stage_Flag = true;                       //Set the Lane1 staged flag
          L2stage_Flag = true;                       //Set the Lane2 staged flag
          Serial.println(SMsg2);                     //Send Staged state message to PC
        }
      }
    }
  }
  Set_Lights(Lane1_lights, Lane2_lights);            //Call subroutine to set Xmas tree lights
  
  //STAGED STATE
  if (racestate == Staged) {                         //Do if timer is in the Staged state
    if (!Hold_Flag) {                                //Disable race start if Hold flag is set
      if (Autostart_Flag) {                          //Do auto start if flag is set
        delay(random(500, 1500));                    //Random wait between 0.5 and 1.5 sec before we start countdown
        racestate = Countdown;                       //Set race state to Countdown state
        Seq_State = 1;                               //Set Sequence State count to 1 (Amber1 on)
        Serial.println(SMsg3);                       //Send Countdown state message to PC
      }
      else {                                         //Do manual start if Autostart_Flag is not set
        Clk_Time = millis();                         //Get the time
        while(digitalRead(hand_sw)==0) {             //Do while hand switch is pressed
          if (millis() > (Clk_Time + 500)) {         //Has switch been held for 1/2 second?
            racestate = Countdown;                   //Set race state to Countdown
            Seq_State = 1;                           //Set Sequence State count to 1 (Amber1 on)
            Serial.println(SMsg3);                   //Send Countdown state message to PC
            break;                                   //exit while loop
          }
          delay(50);                                 //Wait awhile before we loop again
        }
      }
    }
    //The following resets timer to ready state if a car un-stages before the countdown begins
    if (OneRacer_Flag) {                             //Do if only 1 vehicle racing
      if (!Stg1_Flag) {                              //Is Lane1 vehicles nolonger staged?
        racestate = Ready;                           //Set race state back to Ready
        Serial.println(SMsg1);                       //Send Ready state message to PC
      }
    }
    else {                                           //Do if 2 vehicles are racing
      if (!Stg1_Flag || !Stg2_Flag) {                //Is either vehicle nolonger staged?
        racestate = Ready;                           //Set race state back to Ready
        Serial.println(SMsg1);                       //Send Ready state message to PC
      }
    }
  }
  //COUNTDOWN STATE
  //NOTE: Clock time is used extensively when the timer is in the countdown & racing states.
  //      So for speed & efficiency we get the current time just once each pass through the loop
  Clk_Time = millis();                             //Get CPU clock time

  if (racestate == Countdown) {                      //Do if in the Countdown race state
    //TWO RACERS
    if (!OneRacer_Flag) {                            //Do if 2 racers
      if (ProTree_Flag) {                            //Do Pro Xmas Tree lamp sequence if set
        switch (Seq_State) {
          case 1:                                    //Turn on all amber lights
            Seq_Time = Clk_Time + 400;               //Set 400 ms delay for next sequence
            if (Foul_Flag[0]!=1) Lane1_lights = Lane1_lights | AmberAll;  //Set Lane1 Amber Indicators if not 1'st to foul
            if (Foul_Flag[1]!=1) Lane2_lights = Lane2_lights | AmberAll;  //Set Lane2 Amber Indicators if not 1'st to foul
            Seq_State = 2;                           //Set Sequence State count to 2 (Green on)
            break;
          case 2:                                    //Turn off amber lights & turn on green light
            if (Clk_Time > Seq_Time) {               //Check if time to turn on Green light
              Lane1_lights = Lane1_lights & ~AmberAll; //Reset all 3 Lane1 Amber Sequence Indicators
              Lane2_lights = Lane2_lights & ~AmberAll; //Reset all 3 Lane2 Amber Sequence Indicators
              if (Foul_Flag[0]!=1) Lane1_lights = Lane1_lights | Green; //Set Lane1 green GO indicator if not 1'st to foul
              if (Foul_Flag[1]!=1) Lane2_lights = Lane2_lights | Green; //Set Lane2 green GO indicator if not 1'st to foul
              Serial.println(SMsg4);                 //Send Racing state message to PC
              Seq_State = 0;                         //Set Sequence State count to 0 (do nothing)
              racestate = Racing;                    //Set race state to Racing
            }
            break;
          default:
            //Do nothing
            break;
        }
      }
      else {                                         //Do standard Xmas Tree lamp sequence
        switch (Seq_State) {
          case 1:                                    //Turn on first amber light
            Seq_Time = Clk_Time + 400;               //Set 0.4 sec delay for next sequence
            if (Foul_Flag[0]!=1) Lane1_lights = Lane1_lights | Amber1; //Set Lane1 Amber1 Indicator if not 1'st to foul
            if (Foul_Flag[1]!=1) Lane2_lights = Lane2_lights | Amber1; //Set Lane2 Amber1 Indicator if not 1'st to foul
            Seq_State = 2;                           //Set Sequence State count to 2 (Amber2 on)
            break;
          case 2:                                    //Turn on second amber light
            if (Clk_Time > Seq_Time) {               //Check if time to turn on Amber2 light
              Lane1_lights = Lane1_lights & ~Amber1; //Reset Lane1 Amber1 Sequence Indicator
              Lane2_lights = Lane2_lights & ~Amber1; //Reset Lane2 Amber1 Sequence Indicator
              if (Foul_Flag[0]!=1) Lane1_lights = Lane1_lights | Amber2; //Set Lane1 Amber2 Indicator if not 1'st to foul
              if (Foul_Flag[1]!=1) Lane2_lights = Lane2_lights | Amber2; //Set Lane2 Amber2 Indicator if not 1'st to foul
              Seq_Time = Seq_Time + 400;             //Add another 0.4 sec delay for next sequence
              Seq_State = 3;                         //Set Sequence State count to 3 (Amber3 on)
            }
            break;
          case 3:                                    //Turn on third amber light
            if (Clk_Time > Seq_Time) {               //Check if time to turn on Amber3 light
              Lane1_lights = Lane1_lights & ~Amber2; //Reset Lane1 Amber2 Sequence Indicator
              Lane2_lights = Lane2_lights & ~Amber2; //Reset Lane2 Amber2 Sequence Indicator
              if (Foul_Flag[0]!=1) Lane1_lights = Lane1_lights | Amber3; //Set Lane1 Amber3 Indicator if not 1'st to foul
              if (Foul_Flag[1]!=1) Lane2_lights = Lane2_lights | Amber3; //Set Lane2 Amber3 Indicator if not 1'st to foul
              Seq_Time = Seq_Time + 400;             //Add another 0.4 sec delay for next sequence
              Seq_State = 4;                         //Set Sequence State count to 4 (Green on)
            }
            break;
          case 4:                                    //Turn on green light
            if (Clk_Time > Seq_Time) {               //Check if time to turn on Green light
              Lane1_lights = Lane1_lights & ~Amber3; //Reset Lane1 Amber3 Sequence Indicator
              Lane2_lights = Lane2_lights & ~Amber3; //Reset Lane2 Amber3 Sequence Indicator
              if (Foul_Flag[0]!=1) Lane1_lights = Lane1_lights | Green; //Set Lane1 green GO indicator if not 1'st to foul
              if (Foul_Flag[0]!=1) Lane2_lights = Lane2_lights | Green; //Set Lane2 green GO indicator if not 1'st to foul
              Serial.println(SMsg4);                 //Send Racing state message to PC
              Seq_State = 0;                         //Set Sequence State count to 0 (do nothing)
              racestate = Racing;                    //Set race state to Racing
            }
            break;
          default:
            //Do nothing
            break;
        }
      }
      if (Foul_Flag[0]==0 && (!Stg1_Flag || Grd1_Flag)) { //Did Lane1 vehicle leave early during countdown?
        React_time[0] = Clk_Time;                    //Get Lane1 reaction time
        Foul_Flag[0] = Foul_Flag[1] + 1;             //Set Lane1 foul flag as 1'st or 2'nd to foul
        if (Foul_Flag[0]==1) Lane1_lights = Lane1_lights | Disq; //Only red-light Lane1 if fouled first (DQ'd)
      }
      if (Foul_Flag[1]==0 && (!Stg2_Flag || Grd2_Flag)) { //Did Lane2 vehicle leave early during countdown?
        React_time[1] = Clk_Time;                    //Get Lane2 reaction time
        Foul_Flag[1] = Foul_Flag[0] + 1;             //Set Lane2 foul flag as 1'st or 2'nd to foul
        if (Foul_Flag[1]==1) Lane2_lights = Lane2_lights | Disq; //Only red-light Lane2 if fouled first (DQ'd)
      }
    }
    if (OneRacer_Flag) {                             //Do if 1 racer
      if (ProTree_Flag) {                            //Do Pro Xmas Tree lamp sequence if set
        switch (Seq_State) {
          case 1:                                    //Turn on all amber lights
            Seq_Time = Clk_Time + 400;               //Set 400 ms delay for next sequence
            Lane1_lights = Lane1_lights | AmberAll;  //Set all 3 Lane1 Amber Sequence Indicators
            Seq_State = 2;                           //Set Sequence State count to 2 (Amber2 on)
            break;
          case 2:                                    //Turn off amber lights & turn on green light
            if (Clk_Time > Seq_Time) {               //Check if time to turn on Green light
              Lane1_lights = Lane1_lights & ~AmberAll; //Reset all 3 Lane1 Amber Sequence Indicators
              if (Foul_Flag[0]!=1) {                 //Turn on Lane1 Green Go Indicator if not DQ'd
                Lane1_lights = Lane1_lights | Green; //Set Lane1 green GO indicator
              }
              Serial.println(SMsg4);                 //Send Racing state message to PC
              Seq_State = 0;                         //Set Sequence State count to 0 (do nothing)
              racestate = Racing;                    //Set race state to Racing
            }
            break;
          default:
            //Do nothing
            break;
        }
      }
      else {                                         //Do standard Xmas Tree lamp sequence
        switch (Seq_State) {
          case 1:                                    //Turn on first amber light
            Seq_Time = Clk_Time + 400;               //Set 0.4 sec delay for next sequence
            Lane1_lights = Lane1_lights | Amber1;    //Set Lane1 Amber1 Sequence Indicator
            Seq_State = 2;                           //Set Sequence State count to 2 (Amber2 on)
            break;
          case 2:                                    //Turn on second amber light
            if (Clk_Time > Seq_Time) {               //Check if time to turn on Amber2 light
              Lane1_lights = Lane1_lights & ~Amber1; //Reset Lane1 Amber1 Sequence Indicator
              Lane1_lights = Lane1_lights | Amber2;  //Set Lane1 Amber2 Sequence Indicator
              Seq_Time = Seq_Time + 400;             //Add another 0.4 sec delay for next sequence
              Seq_State = 3;                         //Set Sequence State count to 3 (Amber3 on)
            }
            break;
          case 3:                                    //Turn on third amber light
            if (Clk_Time > Seq_Time) {               //Check if time to turn on Amber3 light
              Lane1_lights = Lane1_lights & ~Amber2; //Reset Lane1 Amber2 Sequence Indicator
              Lane1_lights = Lane1_lights | Amber3;  //Set Lane1 Amber3 Sequence Indicator
              Seq_Time = Seq_Time + 400;             //Add another 0.4 sec delay for next sequence
              Seq_State = 4;                         //Set Sequence State count to 4 (Green on)
            }
            break;
          case 4:                                    //Turn on green light
            if (Clk_Time > Seq_Time) {               //Check if time to turn on Green light
              Lane1_lights = Lane1_lights & ~Amber3; //Reset Lane1 Amber3 Sequence Indicator
              if (Foul_Flag[0]!=1) {                 //Turn on Lane1 Green Go Indicator if not DQ'd
                Lane1_lights = Lane1_lights | Green; //Set Lane1 green GO indicator
              }
              Serial.println(SMsg4);                 //Send Racing state message to PC
              Seq_State = 0;                         //Set Sequence State count to 0 (do nothing)
              racestate = Racing;                    //Set race state to Racing
            }
            break;
          default:
            //Do nothing
          break;
        }
      }
      if (Foul_Flag[0]==0 && (!Stg1_Flag || Grd1_Flag)) { //Did Lane1 vehicle leave early during countdown?
        React_time[0] = Clk_Time;                    //Get Lane1 reaction time
        Foul_Flag[0] = Foul_Flag[1] + 1;             //Set Lane1 foul flag as 1'st or 2'nd to foul
        if (Foul_Flag[0]==1) Lane1_lights = Lane1_lights | Disq; //Only red-light Lane1 if fouled first (DQ'd)
      }
    }
  }
  Set_Lights(Lane1_lights, Lane2_lights);            //Call subroutine to set Xmas tree lights

  //RACING STATE
  if (racestate == Racing) {                         //Race underway. Check/Wait for everyone to finish
    Start_time = millis();                           //Get start time from CPU clock in milliseconds
    Exp_Time = Start_time + max_duration;            //Set expiration time (race duration) till timeout
    //TWO RACERS
    if (!OneRacer_Flag) {                            //Do if 2 vehicles racing
      while (fincount<Num_Vehicles) {                //Loop until all vehicles have crossed finish line
        Clk_Time = millis();                         //Get the CPU clock time in milliseconds
        if (L1stage_Flag && Foul_Flag[0]==0) {       //Check & skip if Lane1 already tripped the start-line sensor
          if (digitalRead(Grd_Sense1)==Tripped) {    //Is Lane1 Guard sensor tripped?
            React_time[0] = Clk_Time;                //Yes...get time
            Lane1_lights = Lane1_lights & ~Stage;    //Set Lane1 stage light to off
            Set_Lights(Lane1_lights, Lane2_lights);  //Call subroutine to set Xmas tree lights
            L1stage_Flag = false;                    //Set Lane1 as not staged
          }
        }
        if (L2stage_Flag && Foul_Flag[1]==0) {       //Check & skip if Lane2 already tripped the start-line sensor
          if (digitalRead(Grd_Sense2)==Tripped) {    //Is Lane2 Guard sensor tripped?
            React_time[1] = Clk_Time;                //Yes...get time
            Lane2_lights = Lane2_lights & ~Stage;    //Set Lane2 stage light to off
            Set_Lights(Lane1_lights, Lane2_lights);  //Call subroutine to set Xmas tree lights
            L2stage_Flag = false;                    //Set Lane2 as not staged
          }
        }
        if (!Trap1_Flag) {                           //Check & skip if Lane1 already tripped Trap sensor
          if (digitalRead(Trap_Sense1)==Tripped) {   //Is Lane1 trap sensor tripped?
            Trap_time[0] = Clk_Time;                 //Yes...get time
            Trap1_Flag = true;                       //Set Lane1 trap sensed flag
          }
        }
        if (!Trap2_Flag) {                           //Check & skip if Lane2 already tripped Trap sensor
          if (digitalRead(Trap_Sense2)==Tripped) {   //Is Lane2 trap sensor tripped?
            Trap_time[1] = Clk_Time;                 //Yes...get time
            Trap2_Flag = true;                       //Set Lane2 trap sensed flag
          }
        }
        if (lane_place[0]==0) {                      //Check & skip if Lane1 already crossed finish line
          if (digitalRead(Fin_Sense1)==Tripped) {    //Is Car over Lane1 Finish Line?
            Fin_time[0]=Clk_Time;                    //Yes...get time
            lane_place[0]=++fincount;                //Increment finish count & set finish place for Lane1
          }
        }
        if (lane_place[1]==0) {                      //Check & skip if Lane2 already finished
          if (digitalRead(Fin_Sense2)==Tripped) {    //Is Car over Lane2 Finish Line?
            Fin_time[1]=Clk_Time;                    //Yes...get time
            lane_place[1]=++fincount;                //Increment finish count & set finish place for Lane2
          }
        }
        if (millis()>Exp_Time) break;                //Exit 'while' loop if time has expired
      }
    }
    //ONE RACER
    if (OneRacer_Flag) {                             //Do if 1 vehicle racing
      while (fincount<1) {                           //Loop until Lane1 vehicle has crossed finish line
        Clk_Time = millis();                         //Get the CPU clock time in milliseconds
        if (L1stage_Flag && Foul_Flag[0]==0) {       //Check & skip if Lane1 already tripped the start-line sensor
          if (digitalRead(Grd_Sense1)==Tripped) {    //Is Lane1 Guard sensor tripped?
            React_time[0] = Clk_Time;                //Yes...get time
            Lane1_lights = Lane1_lights & ~Stage;    //Set Lane1 stage light to off
            Set_Lights(Lane1_lights, Lane2_lights);  //Call subroutine to set Xmas tree lights
            L1stage_Flag = false;                    //Set Lane1 start-line sensed flag
          }
        }
        if (!Trap1_Flag) {                           //Check & skip if Lane1 already tripped Trap sensor
          if (digitalRead(Trap_Sense1)==Tripped) {   //Is Lane1 trap sensor tripped?
            Trap_time[0] = Clk_Time;                 //Yes...get time
            Trap1_Flag = true;                       //Set Lane1 trap sensed flag
          }
        }
        if (lane_place[0]==0) {                      //Check & skip if Lane1 already crossed finish line
          if (digitalRead(Fin_Sense1)==Tripped) {    //Is Car over Lane1 Finish Line?
            Fin_time[0]=Clk_Time;                    //Yes...get time
            lane_place[0]=++fincount;                //Increment finish count & set finish place for Lane1
          }
        }
        if (millis()>Exp_Time) break;                //Exit 'while' loop if time has expired
      }
    }
    
    //Race is over or time has expired
    racestate=Finished;                              //Set race state to Finished
    Lane1_lights = Lane1_lights & ~Green;            //Set Lane1 green GO light to off
    Lane2_lights = Lane2_lights & ~Green;            //Set Lane2 green GO light to off
    //Calculate times & speed
    for (i=0; i<Num_Vehicles; i++) {                 //Loop through all vehicles
      if (React_time[i]>0) {                         //Check to see if Lane reaction time is valid
        if (Foul_Flag[i]>0) {                        //Guard sensor tripped before start time so time is negative
          Rtime[i] = float((Start_time-React_time[i])/1000.0);  //Calc negative reaction time in seconds
          if (Rtime[i]>=1.0) Rtime[i] = 0.999;       //And limit value to <=0.999 seconds
        }
        else {                                       //Guard sensor tripped after start time so time is positive
          Rtime[i] = float((React_time[i]-Start_time)/1000.0);  //Calc positive reaction time in seconds
        }
      }
      if (Fin_time[i]>0) {                           //Check to see if finish time is valid
        Etime[i] = float(Fin_time[i]-React_time[i])/1000.0; //Calc lane elapsed finish time and convert to seconds
        if (Foul_Flag[i]>0) {                        //Is reaction time negative?
          Otime[i] = Etime[i];                       //Yes - overall time is same as elapsed time
        }
        else {
          Otime[i] = float(Fin_time[i]-Start_time)/1000.0; //Calc lane overall time and convert to seconds
        }
      }
      if (Trap_time[i]>0 && Fin_time[i]>0) {         //Check to see if Trap and finish times are valid
        Xtime = Fin_time[i]-Trap_time[i];            //Calc time between trap & finish (milliseconds)
        if (kmh_Flag) {                              //Do if speed is in km/h
          Speed[i] = float(7200.0/Xtime);            //Calculate speed in km/h (Assumes 2 Meter seperation distance)
        }
        else {                                       //Else do if speed is in mph
          Speed[i] = float(3600.0/Xtime);            //Calculate speed in mph (Assumes 5.28 Ft seperation distance)
        }
      }
    }
    //Send race results to 7-segment displays
    LED_Disp[0].print(Rtime[0],3);                   //Send lane 1 reaction time to display memory w/ 3 decimal accuracy
    if (Foul_Flag[0]>0) {                            //Is Lane1 reaction time negative?
      LED_Disp[0].writeDigitRaw(0, 0xC0);            //Replace leading 0 digit with minus sign & decimal point
    }
    LED_Disp[0].writeDisplay();                      //And write it to the display
    LED_Disp[2].print(Speed[0],2);                   //Send lane 1 speed to display memory w/ 2 decimal accuracy
    LED_Disp[2].writeDisplay();                      //And write it to the display
    LED_Disp[4].print(Etime[0],3);                   //Send lane 1 elapsed time to display memory w/ 3 decimal accuracy
    LED_Disp[4].writeDisplay();                      //And write it to the display
    if (!OneRacer_Flag) {                            //Do if 2 racers
      LED_Disp[1].print(Rtime[1],3);                 //Send lane 2 reaction time to display memory w/ 3 decimal accuracy
      if (Foul_Flag[1]>0) {                          //Is Lane2 reaction time negative?
        LED_Disp[1].writeDigitRaw(0, 0xC0);          //Replace leading 0 digit with minus sign & decimal point
      }
      LED_Disp[1].writeDisplay();                    //And write it to the display
      LED_Disp[3].print(Speed[1],2);                 //Send lane 2 speed to display memory w/ 2 decimal accuracy
      LED_Disp[3].writeDisplay();                    //And write it to the display
      LED_Disp[5].print(Etime[1],3);                 //Send lane 2 elapsed time to display memory w/ 3 decimal accuracy
      LED_Disp[5].writeDisplay();                    //And write it to the display
    }
    else {                                           //Do if 1 racer
      Set_Lane_Dashes(1);                            //Write dashes to Lane2 RT display
      Set_Lane_Dashes(3);                            //Write dashes to Lane2 speed display
      Set_Lane_Dashes(5);                            //Write dashes to Lane2 ET display
    }
    //Determine winner
    if (!OneRacer_Flag) {                            //Do if 2 vehicles racing
      if (Fin_time[0]>0 && Fin_time[1]>0) {          //If both lanes finished
        if (Foul_Flag[0]==0 && Foul_Flag[1]==0) {    //If both lanes didn't DQ
          if (Otime[0]<Otime[1]) {                   //If lane 1 finished first
            L1blink_Flag=true;                       //Lane1 is winner. Blink Lane1 win light
            blinkTime = millis() + 15000;            //Set 15 second blink time
          }
          if (Otime[0]>Otime[1]) {                   //If lane 2 finished first
            L2blink_Flag=true;                       //Lane2 is winner. Blink Lane2 win light
            blinkTime = millis() + 15000;            //Set 15 second blink time
          }
          if (Otime[0]==Otime[1]) {                  //If both lanes tied
            //tbd - draw
          }
        }
        if (Foul_Flag[0]==1 && Foul_Flag[1]==0) {    //If lane 1 only DQ'd
          L2blink_Flag=true;                         //Lane2 is winner. Blink Lane2 win light
          blinkTime = millis() + 15000;              //Set 15 second blink time
        }
        if (Foul_Flag[0]==0 && Foul_Flag[1]==1) {    //If lane 2 only DQ'd
          L1blink_Flag=true;                         //Lane1 is winner. Blink Lane1 win light
          blinkTime = millis() + 15000;              //Set 15 second blink time
        }
        if (Foul_Flag[0]>0 && Foul_Flag[1]>0) {      //If both lanes DQ'd
          //The one who DQ'd first is the loser. (i.e. One with shortest negative React time is winner?
          //Note: If both cars DQ only the first DQ light will illuminate.
          if (Foul_Flag[0]==1 && Foul_Flag[1]==2) {  //Did Lane1 DQ first?
            L2blink_Flag=true;                       //Lane2 is winner. Blink Lane2 win light
            blinkTime = millis() + 15000;            //Set 15 second blink time
          }
          if (Foul_Flag[0]==2 && Foul_Flag[1]==1) {  //Did Lane2 DQ first?
            L1blink_Flag=true;                       //Lane1 is winner. Blink Lane1 win light
            blinkTime = millis() + 15000;            //Set 15 second blink time
          }
        }
      }
      if (Fin_time[0]>0 && Fin_time[1]==0) {         //If only lane 1 finished
        if (Foul_Flag[0]==0) {                       //If lane 1 did not DQ
          L1blink_Flag=true;                         //Lane1 is winner. Blink Lane1 win light
          blinkTime = millis() + 15000;              //Set 15 second blink time
        }
      }
      if (Fin_time[0]==0 && Fin_time[1]>0) {         //If only lane 2 finished
        if (Foul_Flag[1]==0) {                       //If lane 2 did not DQ
          L2blink_Flag=true;                         //Lane2 is winner. Blink Lane2 win light
          blinkTime = millis() + 15000;              //Set 15 second blink time
        }
      }
      if (Fin_time[0]==0 && Fin_time[1]==0) {        //If neither lane finished
        //No winner
      }
    }
    else {                                           //Do if only one racer (Lane1)
      if (Fin_time[0]>0 && Foul_Flag[0]==0) {        //If lane 1 finished and didn't DQ
        L1blink_Flag=true;                           //Lane1 is winner by default. Blink win light
        blinkTime = millis() + 15000;                //Set 15 second blink time
      }
    }
    if (L1blink_Flag) Win_Flag[0] = true;
    if (L2blink_Flag) Win_Flag[1] = true;
    
    //Send Race Results To PC as one long string for each lane
    //Format: Lanex: y aaaa.aaa bbbb.bbb cccc.ccc dddd.ddd w
    //Where: x=lane#, y=DQ flag, aaaa.aaa=Reaction time, bbbb.bbb=Trap time, cccc.ccc=Finish time,
    //       dddd.ddd=Speed, w=Win flag
    for (i=0; i<2; i++) {                            //Loop through all lanes
      Serial.print("Lane");                          //Precede string with "Lane" as an identifier
      Serial.print(i+1);                             //Print lane number
      Serial.print(": ");                            //Add a space to separate numbers
      if (Foul_Flag[i]==0) Serial.print("0");        //Print lane foul status (0, 1 or 2)
      if (Foul_Flag[i]==1) Serial.print("1");        //Print lane foul status (0, 1 or 2)
      if (Foul_Flag[i]==2) Serial.print("2");        //Print lane foul status (0, 1 or 2)
      Serial.print(" ");                             //Add a space to separate numbers
      if (React_time[i]>0) {                         //Check to see if Lane reaction time is valid
        if (Foul_Flag[i]>0) {                        //Car before green light (negative reaction time)
          Time = float(Start_time-React_time[i])/1000.0; //Calc reaction time in seconds
        }
        else {                                       //Car left after green light (positive reaction time)
          Time = float(React_time[i]-Start_time)/1000.0; //Calc reaction time in seconds
        }
        Time = Time + 1000.0;                        //Add 1000 to ensure number always has XXXX.XXX format
        Serial.print(Time,3);                        //Print lane Reaction time trimmed to 3 places
      }
      else Serial.print("9999.999");                 //Print all nines if lane time is invalid
      Serial.print(" ");                             //Add a space to separate numbers
      if (Trap_time[i]>0) {                          //Check to see if lane Trap time is valid
        Time = float(Trap_time[i]-React_time[i])/1000.0;  //Calc lane Trap time and convert to seconds
        Time = Time + 1000.0;                        //Add 1000 to ensure number always has XXXX.XXX format
        Serial.print(Time,3);                        //Print lane Trap time trimmed to 3 places
      }
      else Serial.print("9999.999");                 //Print all nines if lane Trap time is invalid
      Serial.print(" ");                             //Add a space to separate numbers
      if (Fin_time[i]>0) {                           //Check to see if finish time is valid
        Time = float(Fin_time[i]-React_time[i])/1000.0;  //Calc lane elapsed finish time and convert to seconds
        Time = Time + 1000.0;                        //Add 1000 to ensure number always has XXXX.XXX format
        Serial.print(Time,3);                        //Print lane elapsed finish time trimmed to 3 places after decimal
      }
      else Serial.print("9999.999");                 //Print all nines if lane finish time is invalid
      Serial.print(" ");
      if (Speed[i]>0.0) {                            //Check to see if speed is valid
        Temp = Speed[i] + 1000.0;                    //Add 1000 to ensure number always has XXXX.XXX format
        Serial.print(Temp,3);                        //Print lane speed trimmed to 3 places
      }
      else Serial.print("9999.999");                 //Print all nines if speed is invalid
      Serial.print(" ");
      Serial.println(Win_Flag[i]);                   //Send lane win status & cr/lf to indicate end of the string
      delay(100);                                    //Wait a while to allow display SW to process serial data
    }
    Serial.println(SMsg5);                           //Send Finish state message to PC
  }                                                  //End Racing "if"
  
  //Blink Green "Go" Indicator to show winner
  if (L1blink_Flag) {                                //Check if Lane1 blink flag is set
    Xtime = millis() & 512;                          //Mask time with binary bit that toggles every 512 ms
    if (Xtime==512) {
      Lane1_lights = Lane1_lights | Green;           //Set Lane1 "Go" light to on
    }
    else {
      Lane1_lights = Lane1_lights & ~Green;          //Set Lane1 "Go" light to off
      if (millis()>blinkTime) L1blink_Flag=false;    //Force end if blink time expired
    }
  }
  if (L2blink_Flag) {                                //Check if Lane2 blink flag is set
    Xtime = millis() & 512;                          //Mask time with binary bit that toggles every 512 ms
    if (Xtime==512) {
      Lane2_lights = Lane2_lights | Green;           //Set Lane2 "Go" light to on
    }
    else {
      Lane2_lights = Lane2_lights & ~Green;          //Set Lane2 "Go" light to off
      if (millis()>blinkTime) L2blink_Flag=false;    //Force end if blink time expired
    }
  }

  //FINISHED STATE
  //Manual pushbutton reset routine. Pushbutton must be held for 0.25 seconds to cause a reset
  if (racestate==Finished || racestate==NotReady || racestate==Ready) { //If not staged or racing
    Clk_Time = millis();                             //Get the time
    while(digitalRead(hand_sw)==0) {                 //Do while hand switch is pressed
      if (millis() > (Clk_Time + 250)) {             //Has switch been held for 1/4 second?
        Serial.println(SMsg8);                       //Send reset request message to PC
        timer_reset();                               //Call timer reset routine
        break;                                       //exit while loop
      }
      delay(50);                                     //Wait awhile before we loop again
    }
  }
}                                                    //Loop back


//===================================== SUBROUTINES & FUNCTIONS =====================================
//Subroutine: Initialize Racing Variables
void initialize() {  
  for (int n=0; n<2; n++) {                        //Reset lane Place, Times
    lane_place[n] = 0;
    React_time[n] = 0.0;
    Trap_time[n] = 0.0;
    Fin_time[n] = 0.0;
    Rtime[n] = 0.0;
    Etime[n] = 0.0;
    Otime[n] = 0.0;
    Speed[n] = 0.0;
  }
  Hold_Flag = false;                               //Clear Hold flag
  L1stage_Flag = false;                            //Reset Lane1 start-line sense flag
  L2stage_Flag = false;                            //Reset Lane2 start-line sense flag
  Stg1_Flag = false;                               //Reset Lane1 stage sense flag
  Stg2_Flag = false;                               //Reset Lane2 stage sense flag
  Grd1_Flag = false;                               //Reset Lane1 guard sense flag
  Grd2_Flag = false;                               //Reset Lane2 guard sense flag
  Trap1_Flag = false;                              //Reset Lane1 Trap sense flag
  Trap2_Flag = false;                              //Reset Lane2 Trap sense flag
  fincount = 0;                                    //Reset finish count to 0
  Serial.flush();                                  //Clear the serial input buffer
  for (n=0; n<6; n++) {                            //Loop thru all 6 displays
    LED_Disp[n].clear();                           //Send clear command to display memory
    LED_Disp[n].writeDisplay();                    //And clear the display
  }
  return;
}

//Subroutine: Reset Timer
void timer_reset() {
  Foul_Flag[0] = 0;                                //Set Lane1 disqualification flag to false
  Foul_Flag[1] = 0;                                //Set Lane2 disqualification flag to false
  Win_Flag[0] = false;                             //Set Lane1 win flag to false
  Win_Flag[1] = false;                             //Set Lane2 win flag to false
  L1blink_Flag = false;                            //Set Lane1 blink command to false (off)
  L2blink_Flag = false;                            //Set Lane2 blink command to false (off)
  blinkTime = 0;                                   //Reset blink time
  Xtime = 0;                                       //Clear time variable
  Lane1_lights = 0x00;                             //Set all Lane1 lights off
  Lane2_lights = 0x00;                             //Set all Lane2 lights off
  Set_Lights(Lane1_lights, Lane2_lights);          //Call subroutine to set Xmas tree lights
  Check_Track();                                   //Call Check Track subroutine
  if (Track_ready){                                //and only do a reset if Track is ready
    initialize();                                  //Call initialize function & reset everything
    racestate = Ready;                             //Set race state to Ready
    Serial.println(SMsg1);                         //Send Ready state message to PC
  } 
  else {                                           //Do this if track is not ready
    racestate = NotReady;                          //Set race state to Not Ready
    Serial.println(SMsg0);                         //Send Not Ready state message to PC
  }
}

//Function: Read Serial Data
String get_serial_data() {
  String Str = "";
  if (Serial.available() > 0) {
    Str = Serial.readStringUntil('\n');
    Serial.flush();
    return Str;
  }
}  

//Subroutine: Check Track - Sets 'Track_ready' to true or false
void Check_Track() {
  Track_ready = true;                              //Default track ready status to true
  Lane1_ready = true;                              //Default Lane1 sensor ready status to true
  Lane2_ready = true;                              //Default Lane2 sensor ready status to true
  error_codes = "";                                //Clear bad sensor string
  //Check Lane1 Sensors
  if (digitalRead(Stg_Sense1)==Tripped) {          //Check if a Lane1 Stage sensor is obstructed
    Lane1_ready = false;                           //If so, set Lane1 sensor ready status to false
    error_codes = error_codes + " Stg1";           //Build error string identifying obstructed sensors
  }
  if (digitalRead(Grd_Sense1)==Tripped) {          //Check if a Lane1 Guard sensor is obstructed
    Lane1_ready = false;                           //If so, set Lane1 sensor ready status to false
    error_codes = error_codes + " Grd1";           //Build error string identifying obstructed sensors
  }
  if (digitalRead(Trap_Sense1)==Tripped) {         //Check if a Lane1 trap sensor 1s obstructed
    Lane1_ready = false;                           //If so, set Lane1 sensor ready status to false
    error_codes = error_codes + " Trp1";           //Build error string identifying obstructed lane sensors
  }
  if (digitalRead(Fin_Sense1)==Tripped) {          //Check if a Lane1 finish sensor is obstructed
    Lane1_ready = false;                           //If so, set Lane1 sensor ready status to false
    error_codes = error_codes + " Fin1";           //Build error string identifying obstructed lane sensors
  }
  //Check Lane2 Sensors
  if (digitalRead(Stg_Sense2)==Tripped) {          //Check if a Lane2 start-line sensor is obstructed
    Lane2_ready = false;                           //If so, set Lane2 sensor ready status to false
    error_codes = error_codes + " Stg2";           //Build error string identifying obstructed sensors
  }
  if (digitalRead(Grd_Sense2)==Tripped) {          //Check if a Lane2 Guard sensor is obstructed
    Lane2_ready = false;                           //If so, set Lane2 sensor ready status to false
    error_codes = error_codes + " Grd2";           //Build error string identifying obstructed sensors
  }
  if (digitalRead(Trap_Sense2)==Tripped) {         //Check if a Lane2 trap sensor 1s obstructed
    Lane2_ready = false;                           //If so, set Lane2 sensor ready status to false
    error_codes = error_codes + " Trp2";           //Build error string identifying obstructed lane sensors
  }
  if (digitalRead(Fin_Sense2)==Tripped) {          //Check if a Lane2 finish sensor is obstructed
    Lane2_ready = false;                           //If so, set Lane2 sensor ready status to false
    error_codes = error_codes + " Fin2";           //Build error string identifying obstructed lane sensors
  }
  if (!Lane1_ready || !Lane2_ready) {              //Check if either lane is not ready
    Serial.println(SMsg7 + error_codes);           //If so, send error message to PC
    Track_ready = false;                           //Set track ready status to false
  }
  for (i=0; i<2; i++) {                            //And do lane ready or not_ready blink routine
    if (Lane1_ready) Lane1_lights = Green;         //Set green light if lane 1 ready
      else Lane1_lights = Disq;                    //Set red DQ light if lane 1 not ready
    
    if (Lane2_ready) Lane2_lights = Green;         //Set green light if lane 2 ready
      else Lane2_lights = Disq;                    //Set red DQ light if lane 2 not ready
    
    Set_Lights(Lane1_lights, Lane2_lights);        //Call subroutine to set Xmas tree lights
    delay(75);
    Lane1_lights = 0x00;                           //Set Lane1 lights to off
    Lane2_lights = 0x00;                           //Set Lane2 lights to off
    Set_Lights(Lane1_lights, Lane2_lights);        //Call subroutine to set Xmas tree lights
    delay(200);
  }
}

//Subroutine - Xmas tree & display lamp test
void Lamptest() {
  Lane1_lights = Lane2_lights = Stage;             //Set Lane1 & Lane2 Stage light
  Set_Lights(Lane1_lights, Lane2_lights);          //Call subroutine to send data
  delay(500);                                      //Wait 0.5 seconds
  Lane1_lights = Lane2_lights = Amber1;            //Set Lane1 & Lane2 Amber1 light
  Set_Lights(Lane1_lights, Lane2_lights);          //Call subroutine to send data
  delay(500);                                      //Wait 0.5 seconds
  Lane1_lights = Lane2_lights = Amber2;            //Set Lane1 & Lane2 Amber2 light
  Set_Lights(Lane1_lights, Lane2_lights);          //Call subroutine to send data
  delay(500);                                      //Wait 0.5 seconds
  Lane1_lights = Lane2_lights = Amber3;            //Set Lane1 & Lane2 Amber3 light
  Set_Lights(Lane1_lights, Lane2_lights);          //Call subroutine to send data
  delay(500);                                      //Wait 0.5 seconds
  Lane1_lights = Lane2_lights = Green;             //Set Lane1 & Lane2 Green light
  Set_Lights(Lane1_lights, Lane2_lights);          //Call subroutine to send data
  delay(500);                                      //Wait 0.5 seconds
  Lane1_lights = Lane2_lights = Disq;              //Set Lane1 & Lane2 Disqualification light
  Set_Lights(Lane1_lights, Lane2_lights);          //Call subroutine to send data
  delay(500);                                      //Wait 0.5 seconds
  Lane1_lights = Lane2_lights = 0x00;              //Set Lane1 & Lane2 lights to off
  Set_Lights(Lane1_lights, Lane2_lights);          //Call subroutine to send data
  for (i=1; i<10; i++) {                           //Loop to generate test values 0.000 thru 9.999
    Time = float(i * 1.111);                       //Derive numeric test pattern
    for (n=0; n<6; n++) {                          //Loop thru all 6 displays
      LED_Disp[n].print(Time,3);                   //Send to display memory w/ 3 decimal accuracy
      LED_Disp[n].writeDisplay();                  //And write it to the display
    }
    delay(500);                                    //Cycle at 2 Hz rate
  }
  for (n=0; n<6; n++) {                            //Loop thru all 6 displays
    LED_Disp[n].clear();                           //Send clear command to display memory
    LED_Disp[n].writeDisplay();                    //And clear the display
  }
}

//Subroutine - Send serial data to XMAS Tree LED data registers
byte Set_Lights(byte L1byte, byte L2byte) {
  SPI.transfer(L2byte);                            //Serially send Lane2 data LSB first
  SPI.transfer(L1byte);                            //Serially send Lane1 data LSB first
  digitalWrite(OutReg_CLK, HIGH);                  //Togle the output register clock
  digitalWrite(OutReg_CLK, LOW);                   //to latch in data
}

// Subroutine: Turn on selected Lane Dashes. Enter with Display Select
int Set_Lane_Dashes(int Disp_Select) {
  for (int digit=0; digit<4; digit++) {            //Loop thru all 4 digits
    LED_Disp[Disp_Select].writeDigitRaw(digit+int(digit/2), 0x40);  //And set dash (segment G) for each
  }
  LED_Disp[Disp_Select].drawColon(false);          //Set colon command to false just in case it's true
  LED_Disp[Disp_Select].writeDisplay();            //And send above commands to the display
}
